@@ -427,6 +427,8 @@ var parseTimeZoneTests = []ParseTimeZoneTest{
 	{"ESASTT hi", 0, false}, // run of upper-case letters too long.
 	{"ESATY hi", 0, false},  // five letters must end in T.
 	{"WITA hi", 4, true},    // Issue #18251
+	{"+03 hi", 3, true},     // Issue #24071
+	{"-04 hi", 3, true},     // Issue #24071
 }
 
 func TestParseTimeZone(t *testing.T) {
